//! Additional documentation: recipes, specific use cases and examples, etc.

#[doc = include_str!("../../doc/RECIPES.md")]
pub mod recipes {}

#[doc = include_str!("../../doc/DERIVE.md")]
pub mod derive {}
