#[derive(PartialEq, Eq, Copy, Clone, Debug)]
#[repr(C)]
pub enum BrotliEncoderParameter {
  BROTLI_PARAM_MODE = 0,
  BROTLI_PARAM_QUALITY = 1,
  BROTLI_PARAM_LGWIN = 2,
  BROTLI_PARAM_LGBLOCK = 3,
  BROTLI_PARAM_DISABLE_LITERAL_CONTEXT_MODELING = 4,
  BROTLI_PARAM_SIZE_HINT = 5,
  BROTLI_PARAM_LARGE_WINDOW = 6,
  BROTLI_PARAM_Q9_5 = 150,
  BROTLI_METABLOCK_CALLBACK = 151,
  BROTLI_PARAM_STRIDE_DETECTION_QUALITY = 152,
  BROTLI_PARAM_HIGH_ENTROPY_DETECTION_QUALITY = 153,
  BROTLI_PARAM_LITERAL_BYTE_SCORE = 154,
  BROTLI_PARAM_CDF_ADAPTATION_DETECTION = 155,
  BROTLI_PARAM_PRIOR_BITMASK_DETECTION = 156,
  BROTLI_PARAM_SPEED = 157,
  BROTLI_PARAM_SPEED_MAX = 158,
  BROTLI_PARAM_CM_SPEED = 159,
  BROTLI_PARAM_CM_SPEED_MAX = 160,
  BROTLI_PARAM_SPEED_LOW = 161,
  BROTLI_PARAM_SPEED_LOW_MAX = 162,
  BROTLI_PARAM_CM_SPEED_LOW = 164,
  BROTLI_PARAM_CM_SPEED_LOW_MAX = 165,
  BROTLI_PARAM_AVOID_DISTANCE_PREFIX_SEARCH = 166,
  BROTLI_PARAM_CATABLE = 167,
  BROTLI_PARAM_APPENDABLE = 168,
  BROTLI_PARAM_MAGIC_NUMBER = 169,
  BROTLI_PARAM_NO_DICTIONARY = 170,
  BROTLI_PARAM_FAVOR_EFFICIENCY = 171,
  UNUSED7=7,
  UNUSED8=8,
  UNUSED9=9,
  UNUSED10=10,
  UNUSED11=11,
  UNUSED12=12,
  UNUSED13=13,
  UNUSED14=14,
  UNUSED15=15,
  UNUSED16=16,
  UNUSED17=17,
  UNUSED18=18,
  UNUSED19=19,
  UNUSED20=20,
  UNUSED21=21,
  UNUSED22=22,
  UNUSED23=23,
  UNUSED24=24,
  UNUSED25=25,
  UNUSED26=26,
  UNUSED27=27,
  UNUSED28=28,
  UNUSED29=29,
  UNUSED30=30,
  UNUSED31=31,
  UNUSED32=32,
  UNUSED33=33,
  UNUSED34=34,
  UNUSED35=35,
  UNUSED36=36,
  UNUSED37=37,
  UNUSED38=38,
  UNUSED39=39,
  UNUSED40=40,
  UNUSED41=41,
  UNUSED42=42,
  UNUSED43=43,
  UNUSED44=44,
  UNUSED45=45,
  UNUSED46=46,
  UNUSED47=47,
  UNUSED48=48,
  UNUSED49=49,
  UNUSED50=50,
  UNUSED51=51,
  UNUSED52=52,
  UNUSED53=53,
  UNUSED54=54,
  UNUSED55=55,
  UNUSED56=56,
  UNUSED57=57,
  UNUSED58=58,
  UNUSED59=59,
  UNUSED60=60,
  UNUSED61=61,
  UNUSED62=62,
  UNUSED63=63,
  UNUSED64=64,
  UNUSED65=65,
  UNUSED66=66,
  UNUSED67=67,
  UNUSED68=68,
  UNUSED69=69,
  UNUSED70=70,
  UNUSED71=71,
  UNUSED72=72,
  UNUSED73=73,
  UNUSED74=74,
  UNUSED75=75,
  UNUSED76=76,
  UNUSED77=77,
  UNUSED78=78,
  UNUSED79=79,
  UNUSED80=80,
  UNUSED81=81,
  UNUSED82=82,
  UNUSED83=83,
  UNUSED84=84,
  UNUSED85=85,
  UNUSED86=86,
  UNUSED87=87,
  UNUSED88=88,
  UNUSED89=89,
  UNUSED90=90,
  UNUSED91=91,
  UNUSED92=92,
  UNUSED93=93,
  UNUSED94=94,
  UNUSED95=95,
  UNUSED96=96,
  UNUSED97=97,
  UNUSED98=98,
  UNUSED99=99,
  UNUSED100=100,
  UNUSED101=101,
  UNUSED102=102,
  UNUSED103=103,
  UNUSED104=104,
  UNUSED105=105,
  UNUSED106=106,
  UNUSED107=107,
  UNUSED108=108,
  UNUSED109=109,
  UNUSED110=110,
  UNUSED111=111,
  UNUSED112=112,
  UNUSED113=113,
  UNUSED114=114,
  UNUSED115=115,
  UNUSED116=116,
  UNUSED117=117,
  UNUSED118=118,
  UNUSED119=119,
  UNUSED120=120,
  UNUSED121=121,
  UNUSED122=122,
  UNUSED123=123,
  UNUSED124=124,
  UNUSED125=125,
  UNUSED126=126,
  UNUSED127=127,
  UNUSED128=128,
  UNUSED129=129,
  UNUSED130=130,
  UNUSED131=131,
  UNUSED132=132,
  UNUSED133=133,
  UNUSED134=134,
  UNUSED135=135,
  UNUSED136=136,
  UNUSED137=137,
  UNUSED138=138,
  UNUSED139=139,
  UNUSED140=140,
  UNUSED141=141,
  UNUSED142=142,
  UNUSED143=143,
  UNUSED144=144,
  UNUSED145=145,
  UNUSED146=146,
  UNUSED147=147,
  UNUSED148=148,
  UNUSED149=149,
  UNUSED172=172,
  UNUSED173=173,
  UNUSED174=174,
  UNUSED175=175,
  UNUSED176=176,
  UNUSED177=177,
  UNUSED178=178,
  UNUSED179=179,
  UNUSED180=180,
  UNUSED181=181,
  UNUSED182=182,
  UNUSED183=183,
  UNUSED184=184,
  UNUSED185=185,
  UNUSED186=186,
  UNUSED187=187,
  UNUSED188=188,
  UNUSED189=189,
  UNUSED190=190,
  UNUSED191=191,
  UNUSED192=192,
  UNUSED193=193,
  UNUSED194=194,
  UNUSED195=195,
  UNUSED196=196,
  UNUSED197=197,
  UNUSED198=198,
  UNUSED199=199,
  UNUSED200=200,
  UNUSED201=201,
  UNUSED202=202,
  UNUSED203=203,
  UNUSED204=204,
  UNUSED205=205,
  UNUSED206=206,
  UNUSED207=207,
  UNUSED208=208,
  UNUSED209=209,
  UNUSED210=210,
  UNUSED211=211,
  UNUSED212=212,
  UNUSED213=213,
  UNUSED214=214,
  UNUSED215=215,
  UNUSED216=216,
  UNUSED217=217,
  UNUSED218=218,
  UNUSED219=219,
  UNUSED220=220,
  UNUSED221=221,
  UNUSED222=222,
  UNUSED223=223,
  UNUSED224=224,
  UNUSED225=225,
  UNUSED226=226,
  UNUSED227=227,
  UNUSED228=228,
  UNUSED229=229,
  UNUSED230=230,
  UNUSED231=231,
  UNUSED232=232,
  UNUSED233=233,
  UNUSED234=234,
  UNUSED235=235,
  UNUSED236=236,
  UNUSED237=237,
  UNUSED238=238,
  UNUSED239=239,
  UNUSED240=240,
  UNUSED241=241,
  UNUSED242=242,
  UNUSED243=243,
  UNUSED244=244,
  UNUSED245=245,
  UNUSED246=246,
  UNUSED247=247,
  UNUSED248=248,
  UNUSED249=249,
  UNUSED250=250,
  UNUSED251=251,
  UNUSED252=252,
  UNUSED253=253,
  UNUSED254=254,
  UNUSED255=255,
}
