//! Counter Mode with a 32-bit counter.
//!
//! NOTE: AES-128-CTR test vectors used by these tests were generated by first
//! integration testing the implementation in the contexts of AES-GCM and
//! AES-GCM-SIV, with the former tested against the NIST CAVS vectors, and the
//! latter against the RFC8452 test vectors.

mod big_endian;
mod little_endian;
